@@ -18,6 +18,7 @@ from dipdup.config import (
     DatasourceConfigT,
     DipDupConfig,
     HandlerConfig,
+    HeadIndexConfig,
     IndexTemplateConfig,
     OperationHandlerOriginationPatternConfig,
     OperationHandlerTransactionPatternConfig,
@@ -96,14 +97,14 @@ class DipDupCodeGenerator:
 
     async def create_package(self) -> None:
         """Create Python package skeleton if not exists"""
-        self._logger.info('Creating package `%s`', self._config.package)
         try:
             package_path = self._config.package_path
         except ImportError:
+            self._logger.info('Creating package `%s`', self._config.package)
             package_path = join(os.getcwd(), self._config.package)
-            touch(join(package_path, '__init__.py'))
 
-        self._logger.info('Creating `%s.models` module', self._config.package)
+        touch(join(package_path, '__init__.py'))
+
         models_path = join(package_path, 'models.py')
         if not exists(models_path):
             template = load_template('models.py')
@@ -111,16 +112,12 @@ class DipDupCodeGenerator:
             write(models_path, models_code)
 
         for subpackage in ('handlers', 'hooks'):
-            self._logger.info('Creating `%s.%s` package', self._config.package, subpackage)
             subpackage_path = join(self._config.package_path, subpackage)
             touch(join(subpackage_path, '__init__.py'))
 
-        self._logger.info('Creating `%s/sql` directory', self._config.package)
         sql_path = join(self._config.package_path, 'sql')
-        touch(join(sql_path, 'on_restart', '.keep'))
-        touch(join(sql_path, 'on_reindex', '.keep'))
+        touch(join(sql_path, '.keep'))
 
-        self._logger.info('Creating `%s/graphql` directory', self._config.package)
         graphql_path = join(self._config.package_path, 'graphql')
         touch(join(graphql_path, '.keep'))
 
@@ -211,6 +208,9 @@ class DipDupCodeGenerator:
                     big_map_value_schema_path = join(big_map_schemas_path, f'{big_map_path}_value.json')
                     write(big_map_value_schema_path, json.dumps(big_map_value_schema, indent=4))
 
+            elif isinstance(index_config, HeadIndexConfig):
+                pass
+
             elif isinstance(index_config, IndexTemplateConfig):
                 raise ConfigInitializationException
 
@@ -276,11 +276,7 @@ class DipDupCodeGenerator:
         """Generate handler stubs with typehints from templates if not exist"""
         handler_config: HandlerConfig
         for index_config in self._config.indexes.values():
-            if isinstance(index_config, OperationIndexConfig):
-                for handler_config in index_config.handlers:
-                    await self._generate_callback(handler_config)
-
-            elif isinstance(index_config, BigMapIndexConfig):
+            if isinstance(index_config, (OperationIndexConfig, BigMapIndexConfig, HeadIndexConfig)):
                 for handler_config in index_config.handlers:
                     await self._generate_callback(handler_config)
 
