@@ -601,23 +601,35 @@ class IndexConfig(TemplateValuesMixin, NameMixin, ParentMixin['ResolvedIndexConf
         NameMixin.__post_init_post_parse__(self)
         ParentMixin.__post_init_post_parse__(self)
 
-    def dumps(self) -> str:
-        return json.dumps(self, default=pydantic_encoder)
-
-    def json(self) -> Dict[str, Any]:
-        return json.loads(self.dumps())
-
-    def hash(self) -> str:
-        config_json = self.dumps()
-        config_hash = hashlib.sha256(config_json.encode()).hexdigest()
-        return config_hash
-
     @property
     def datasource_config(self) -> TzktDatasourceConfig:
         if not isinstance(self.datasource, TzktDatasourceConfig):
             raise ConfigInitializationException
         return self.datasource
 
+    def hash(self) -> str:
+        """Calculate hash to ensure config not changed since last run."""
+        config_json = json.dumps(self, default=pydantic_encoder)
+
+        # FIXME: How to convert pydantic dataclass into dict without json.dumps? asdict is not recursive.
+        config_dict = json.loads(config_json)
+
+        # NOTE: We need to preserve datasource URL but remove it's HTTP tunables to avoid false-positives.
+        config_dict['datasource'].pop('http', None)
+
+        config_json = json.dumps(config_dict)
+        config_hash = hashlib.sha256(config_json.encode()).hexdigest()
+        return config_hash
+
+    def hash_old(self) -> str:
+        """Calculate hash to ensure config not changed since last run.
+
+        Old incorrect algorightm (false positives). Used only to update hash of existing indexes.
+        """
+        config_json = json.dumps(self, default=pydantic_encoder)
+        config_hash = hashlib.sha256(config_json.encode()).hexdigest()
+        return config_hash
+
 
 @dataclass
 class OperationIndexConfig(IndexConfig):
@@ -750,8 +762,28 @@ class BigMapIndexConfig(IndexConfig):
         return list(set([cast(ContractConfig, handler_config.contract) for handler_config in self.handlers]))
 
 
-IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, IndexTemplateConfig]
-ResolvedIndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig]
+@dataclass
+class HeadHandlerConfig(HandlerConfig, kind='handler'):
+    def iter_imports(self, package: str) -> Iterator[Tuple[str, str]]:
+        yield 'dipdup.context', 'HandlerContext'
+        yield 'dipdup.models', 'HeadBlockData'
+        yield package, 'models as models'
+
+    def iter_arguments(self) -> Iterator[Tuple[str, str]]:
+        yield 'ctx', 'HandlerContext'
+        yield 'head', 'HeadBlockData'
+
+
+@dataclass
+class HeadIndexConfig(IndexConfig):
+    kind: Literal['head']
+    datasource: Union[str, TzktDatasourceConfig]
+    handlers: List[HeadHandlerConfig]
+
+
+IndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, HeadIndexConfig, IndexTemplateConfig]
+ResolvedIndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig, HeadIndexConfig]
+ContractIndexConfigT = Union[OperationIndexConfig, BigMapIndexConfig]
 HandlerPatternConfigT = Union[OperationHandlerOriginationPatternConfig, OperationHandlerTransactionPatternConfig]
 
 
@@ -968,6 +1000,12 @@ class DipDupConfig:
         except KeyError as e:
             raise ConfigurationError(f'Datasource `{name}` not found in `datasources` config section') from e
 
+    def get_index(self, name: str) -> IndexConfigT:
+        try:
+            return self.indexes[name]
+        except KeyError as e:
+            raise ConfigurationError(f'Index `{name}` not found in `indexes` config section') from e
+
     def get_template(self, name: str) -> ResolvedIndexConfigT:
         try:
             return self.templates[name]
@@ -1012,6 +1050,9 @@ class DipDupConfig:
                 self._import_big_map_index_types(index_config)
                 self._import_index_callbacks(index_config)
 
+            elif isinstance(index_config, HeadIndexConfig):
+                self._import_index_callbacks(index_config)
+
             else:
                 raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
@@ -1029,8 +1070,13 @@ class DipDupConfig:
             if name in default_hooks:
                 raise ConfigurationError(f'`{name}` hook name is reserved. See docs to learn more about built-in hooks.')
 
-        # NOTE: Duplicate contracts
-        contracts = [cast(ResolvedIndexConfigT, i).contracts for i in self.indexes.values() if cast(ResolvedIndexConfigT, i).contracts]
+        # NOTE: Detect duplicate contracts
+        # FIXME: This code smells
+        contracts = [
+            cast(ContractIndexConfigT, i).contracts
+            for i in self.indexes.values()
+            if hasattr(i, 'contracts') and cast(ContractIndexConfigT, i).contracts
+        ]
         plain_contracts = reduce(operator.add, contracts) if contracts else []  # type: ignore
         # NOTE: After pre_initialize
         duplicate_contracts = [cast(ContractConfig, item).name for item, count in Counter(plain_contracts).items() if count > 1]
@@ -1144,6 +1190,13 @@ class DipDupConfig:
                 if isinstance(handler.contract, str):
                     handler.contract = self.get_contract(handler.contract)
 
+        elif isinstance(index_config, HeadIndexConfig):
+            if isinstance(index_config.datasource, str):
+                index_config.datasource = self.get_tzkt_datasource(index_config.datasource)
+
+            for head_handler_config in index_config.handlers:
+                head_handler_config.parent = index_config
+
         else:
             raise NotImplementedError(f'Index kind `{index_config.kind}` is not supported')
 
