@@ -5,7 +5,7 @@ from collections import deque
 from contextlib import AsyncExitStack, asynccontextmanager, suppress
 from functools import partial
 from operator import ne
-from typing import Awaitable, Deque, Dict, List, Optional, Set
+from typing import Awaitable, Deque, Dict, List, Optional, Set, Tuple, cast
 
 from apscheduler.events import EVENT_JOB_ERROR  # type: ignore
 from tortoise.exceptions import OperationalError
@@ -29,15 +29,15 @@ from dipdup.datasources.coinbase.datasource import CoinbaseDatasource
 from dipdup.datasources.datasource import Datasource, IndexDatasource
 from dipdup.datasources.tzkt.datasource import TzktDatasource
 from dipdup.enums import ReindexingReason
-from dipdup.exceptions import ConfigInitializationException, DipDupException
+from dipdup.exceptions import ConfigInitializationException, DipDupException, ReindexingRequiredError
 from dipdup.hasura import HasuraGateway
-from dipdup.index import BigMapIndex, Index, OperationIndex
+from dipdup.index import BigMapIndex, HeadIndex, Index, OperationIndex, block_cache
 from dipdup.models import BigMapData, Contract, Head, HeadBlockData
 from dipdup.models import Index as IndexState
 from dipdup.models import IndexStatus, OperationData, Schema
 from dipdup.scheduler import add_job, create_scheduler
 from dipdup.utils import slowdown
-from dipdup.utils.database import generate_schema, get_schema_hash, set_schema, tortoise_wrapper, validate_models
+from dipdup.utils.database import generate_schema, get_schema_hash, prepare_models, set_schema, tortoise_wrapper, validate_models
 
 
 class IndexDispatcher:
@@ -64,7 +64,7 @@ class IndexDispatcher:
             while self._tasks:
                 tasks.append(self._tasks.popleft())
 
-            async with slowdown(1.0):
+            async with slowdown(0.1):
                 await gather(*tasks)
 
             indexes_spawned = False
@@ -126,8 +126,19 @@ class IndexDispatcher:
             if index_config := self._ctx.config.indexes.get(name):
                 if isinstance(index_config, IndexTemplateConfig):
                     raise ConfigInitializationException
-                if index_config.hash() != index_state.config_hash:
-                    await self._ctx.reindex(ReindexingReason.CONFIG_HASH_MISMATCH)
+
+                new_hash = index_config.hash()
+                if new_hash != index_state.config_hash:
+                    # NOTE: Try old hashing algorithm. Update if matches, reindex if not.
+                    old_hash = index_config.hash_old()
+                    if old_hash == index_state.config_hash:
+                        self._logger.warning('Updating config hash of index `%s`', name)
+                        index_state.config_hash = new_hash  # type: ignore
+                        await index_state.save()
+                    else:
+                        await self._ctx.reindex(
+                            ReindexingReason.CONFIG_HASH_MISMATCH, index_hash=index_state.config_hash, old_hash=old_hash, new_hash=new_hash
+                        )
 
             # NOTE: Templated index: recreate index config, verify hash
             elif template:
@@ -135,10 +146,13 @@ class IndexDispatcher:
                     await self._ctx.reindex(ReindexingReason.MISSING_INDEX_TEMPLATE)
                 await self._ctx.add_index(name, template, template_values)
 
-            # NOTE: Index config is missing
+            # NOTE: Index config is missing, possibly just commented-out
             else:
                 self._logger.warning('Index `%s` was removed from config, ignoring', name)
 
+        # NOTE: Cached blocks used only on index state init
+        block_cache.clear()
+
     async def _on_head(self, datasource: TzktDatasource, head: HeadBlockData) -> None:
         # NOTE: Do not await query results - blocked database connection may cause Websocket timeout.
         self._tasks.append(
@@ -153,37 +167,52 @@ class IndexDispatcher:
                 ),
             )
         )
+        for index in self._indexes.values():
+            if isinstance(index, HeadIndex) and index.datasource == datasource:
+                index.push_head(head)
 
-    async def _on_operations(self, datasource: TzktDatasource, operations: List[OperationData]) -> None:
-        assert len(set(op.level for op in operations)) == 1
-        level = operations[0].level
+    async def _on_operations(self, datasource: TzktDatasource, operations: Tuple[OperationData, ...]) -> None:
         for index in self._indexes.values():
             if isinstance(index, OperationIndex) and index.datasource == datasource:
-                index.push(level, operations)
+                index.push_operations(operations)
 
-    async def _on_big_maps(self, datasource: TzktDatasource, big_maps: List[BigMapData]) -> None:
-        assert len(set(op.level for op in big_maps)) == 1
-        level = big_maps[0].level
+    async def _on_big_maps(self, datasource: TzktDatasource, big_maps: Tuple[BigMapData]) -> None:
         for index in self._indexes.values():
             if isinstance(index, BigMapIndex) and index.datasource == datasource:
-                index.push(level, big_maps)
+                index.push_big_maps(big_maps)
 
     async def _on_rollback(self, datasource: TzktDatasource, from_level: int, to_level: int) -> None:
-        # NOTE: Rollback could be received before head
-        if from_level - to_level in (0, 1):
-            # NOTE: Single level rollbacks are processed at Index level.
-            # NOTE: Notify all indexes which use rolled back datasource to drop duplicated operations from the next block
-            for index in self._indexes.values():
-                if index.datasource == datasource:
-                    # NOTE: Continue to rollback with handler
-                    if not isinstance(index, OperationIndex):
-                        self._logger.info('Single level rollback is not supported by `%s` indexes', index._config.kind)
-                        break
-                    await index.single_level_rollback(from_level)
-            else:
-                return
+        """Perform a single level rollback when possible, otherwise call `on_rollback` hook"""
+        self._logger.warning('Datasource `%s` rolled back: %s -> %s', datasource.name, from_level, to_level)
 
-        await self._ctx.fire_hook('on_rollback', datasource=datasource, from_level=from_level, to_level=to_level)
+        # NOTE: Zero difference between levels means we received no operations/big_maps on this level and thus channel level hasn't changed
+        zero_level_rollback = from_level - to_level == 0
+        single_level_rollback = from_level - to_level == 1
+
+        if zero_level_rollback:
+            self._logger.info('Zero level rollback, ignoring')
+
+        elif single_level_rollback:
+            # NOTE: Notify all indexes which use rolled back datasource to drop duplicated operations from the next block
+            self._logger.info('Checking if single level rollback is possible')
+            matching_indexes = tuple(i for i in self._indexes.values() if i.datasource == datasource)
+            matching_operation_indexes = tuple(i for i in matching_indexes if isinstance(i, OperationIndex))
+            self._logger.info(
+                'Indexes: %s total, %s matching, %s support single level rollback',
+                len(self._indexes),
+                len(matching_indexes),
+                len(matching_operation_indexes),
+            )
+
+            all_indexes_are_operation = len(matching_indexes) == len(matching_operation_indexes)
+            if all_indexes_are_operation:
+                for index in cast(List[OperationIndex], matching_indexes):
+                    index.push_rollback(from_level)
+            else:
+                await self._ctx.fire_hook('on_rollback', datasource=datasource, from_level=from_level, to_level=to_level)
+
+        else:
+            await self._ctx.fire_hook('on_rollback', datasource=datasource, from_level=from_level, to_level=to_level)
 
 
 class DipDup:
@@ -226,12 +255,12 @@ class DipDup:
     async def docker_init(self, image: str, tag: str, env_file: str) -> None:
         await self._codegen.docker_init(image, tag, env_file)
 
-    async def run(self, reindex: bool, oneshot: bool, postpone_jobs: bool) -> None:
+    async def run(self, oneshot: bool, postpone_jobs: bool) -> None:
         """Run indexing process"""
         tasks: Set[Task] = set()
         async with AsyncExitStack() as stack:
             stack.enter_context(suppress(KeyboardInterrupt, CancelledError))
-            await self._set_up_database(stack, reindex)
+            await self._set_up_database(stack)
             await self._set_up_datasources(stack)
             await self._set_up_hooks()
 
@@ -293,13 +322,24 @@ class DipDup:
             schema_name = self._config.database.schema_name
             await set_schema(conn, schema_name)
 
+        # NOTE: Try to fetch existing schema
         try:
             self._schema = await Schema.get_or_none(name=schema_name)
+
+        # NOTE: No such table yet
         except OperationalError:
             self._schema = None
+
         # TODO: Fix Tortoise ORM to raise more specific exception
         except KeyError:
-            await self._ctx.reindex(ReindexingReason.SCHEMA_HASH_MISMATCH)
+            try:
+                # NOTE: A small migration, ReindexingReason became ReversedEnum
+                for item in ReindexingReason:
+                    await conn.execute_script(f'UPDATE dipdup_schema SET reindex = "{item.name}" WHERE reindex = "{item.value}"')
+
+                self._schema = await Schema.get_or_none(name=schema_name)
+            except KeyError:
+                await self._ctx.reindex(ReindexingReason.SCHEMA_HASH_MISMATCH)
 
         schema_hash = get_schema_hash(conn)
 
@@ -319,10 +359,14 @@ class DipDup:
         elif self._schema.hash != schema_hash:
             await self._ctx.reindex(ReindexingReason.SCHEMA_HASH_MISMATCH)
 
+        elif self._schema.reindex:
+            raise ReindexingRequiredError(self._schema.reindex)
+
         await self._ctx.fire_hook('on_restart')
 
-    async def _set_up_database(self, stack: AsyncExitStack, reindex: bool) -> None:
-        # NOTE: Must be called before Tortoise.init
+    async def _set_up_database(self, stack: AsyncExitStack) -> None:
+        # NOTE: Must be called before entering Tortoise context
+        prepare_models(self._config.package)
         validate_models(self._config.package)
 
         url = self._config.database.connection_string
@@ -330,9 +374,6 @@ class DipDup:
         models = f'{self._config.package}.models'
         await stack.enter_async_context(tortoise_wrapper(url, models, timeout or 60))
 
-        if reindex:
-            await self._ctx.reindex(ReindexingReason.CLI_OPTION)
-
     async def _set_up_hooks(self) -> None:
         for hook_config in default_hooks.values():
             self._ctx.callbacks.register_hook(hook_config)
