@@ -1,8 +1,8 @@
 import textwrap
 import traceback
 from contextlib import contextmanager
-from dataclasses import dataclass
-from typing import Any, Iterator, Optional, Type
+from dataclasses import dataclass, field
+from typing import Any, Dict, Iterator, Optional, Type
 
 import sentry_sdk
 from tabulate import tabulate
@@ -84,7 +84,6 @@ class ConfigurationError(DipDupError):
         """
 
 
-# TODO: Any cases besides model validation?
 @dataclass(frozen=True, repr=False)
 class DatabaseConfigurationError(ConfigurationError):
     """DipDup can't initialize database with given models and parameters"""
@@ -137,17 +136,23 @@ class ReindexingRequiredError(DipDupError):
     """Unable to continue indexing with existing database"""
 
     reason: ReindexingReason
+    context: Dict[str, Any] = field(default_factory=dict)
 
     def _help(self) -> str:
+        additional_context = '\n              '.join(f'{k}: {v}' for k, v in self.context.items())
         return f"""
             Reindexing required!
 
             Reason: {self.reason.value}
 
+            Additional context:
+
+                {additional_context}
+
             You may want to backup database before proceeding. After that perform one of the following actions:
 
-              * Eliminate the cause of reindexing and run `UPDATE dupdup_schema SET reindex = NULL;`
-              * Run `dipdup run --reindex` to truncate database and start indexing from scratch
+                * Eliminate the cause of reindexing and run `dipdup schema approve`
+                * Run `dipdup schema wipe [--immune]` command to drop database and start indexing from scratch
         """
 
 
